@@ -127,7 +127,9 @@ public abstract class WebappClassLoaderBase extends URLClassLoader
     private static final String CLASS_FILE_SUFFIX = ".class";
 
     static {
-        registerAsParallelCapable();
+        if (!JreCompat.isGraalAvailable()) {
+            registerAsParallelCapable();
+        }
         JVM_THREAD_GROUP_NAMES.add(JVM_THREAD_GROUP_SYSTEM);
         JVM_THREAD_GROUP_NAMES.add("RMI Runtime");
     }
@@ -274,7 +276,7 @@ public abstract class WebappClassLoaderBase extends URLClassLoader
     protected boolean delegate = false;
 
 
-    private final HashMap<String,Long> jarModificationTimes = new HashMap<>();
+    private final Map<String,Long> jarModificationTimes = new HashMap<>();
 
 
     /**
@@ -309,15 +311,6 @@ public abstract class WebappClassLoaderBase extends URLClassLoader
     private ClassLoader javaseClassLoader;
 
 
-    /**
-     * All permission.
-     *
-     * @deprecated Unused. This will be removed in Tomcat 9.
-     */
-    @Deprecated
-    protected final Permission allPermission = new java.security.AllPermission();
-
-
     /**
      * Enables the RMI Target memory leak detection to be controlled. This is necessary since the detection can only
      * work on Java 9 if some of the modularity checks are disabled.
@@ -623,12 +616,6 @@ public abstract class WebappClassLoaderBase extends URLClassLoader
 
     // ------------------------------------------------------- Reloader Methods
 
-    /**
-     * Adds the specified class file transformer to this class loader. The transformer will then be able to modify the
-     * bytecode of any classes loaded by this class loader after the invocation of this method.
-     *
-     * @param transformer The transformer to add to the class loader
-     */
     @Override
     public void addTransformer(ClassFileTransformer transformer) {
 
@@ -647,13 +634,6 @@ public abstract class WebappClassLoaderBase extends URLClassLoader
         log.info(sm.getString("webappClassLoader.addTransformer", transformer, getContextName()));
     }
 
-    /**
-     * Removes the specified class file transformer from this class loader. It will no longer be able to modify the byte
-     * code of any classes loaded by the class loader after the invocation of this method. However, any classes already
-     * modified by this transformer will remain transformed.
-     *
-     * @param transformer The transformer to remove
-     */
     @Override
     public void removeTransformer(ClassFileTransformer transformer) {
 
@@ -686,8 +666,8 @@ public abstract class WebappClassLoaderBase extends URLClassLoader
      */
     public boolean modified() {
 
-        if (log.isDebugEnabled()) {
-            log.debug("modified()");
+        if (log.isTraceEnabled()) {
+            log.trace("modified()");
         }
 
         for (Entry<String,ResourceEntry> entry : resourceEntries.entrySet()) {
@@ -778,8 +758,8 @@ public abstract class WebappClassLoaderBase extends URLClassLoader
     @Override
     public Class<?> findClass(String name) throws ClassNotFoundException {
 
-        if (log.isDebugEnabled()) {
-            log.debug("    findClass(" + name + ")");
+        if (log.isTraceEnabled()) {
+            log.trace("    findClass(" + name + ")");
         }
 
         checkStateForClassLoading(name);
@@ -839,8 +819,8 @@ public abstract class WebappClassLoaderBase extends URLClassLoader
                 }
             }
             if (clazz == null) {
-                if (log.isDebugEnabled()) {
-                    log.debug("    --> Returning ClassNotFoundException");
+                if (log.isTraceEnabled()) {
+                    log.trace("    --> Returning ClassNotFoundException");
                 }
                 throw new ClassNotFoundException(name);
             }
@@ -853,7 +833,7 @@ public abstract class WebappClassLoaderBase extends URLClassLoader
 
         // Return the class we have located
         if (log.isTraceEnabled()) {
-            log.debug("      Returning class " + clazz);
+            log.trace("      Returning class " + clazz);
         }
 
         if (log.isTraceEnabled()) {
@@ -863,7 +843,7 @@ public abstract class WebappClassLoaderBase extends URLClassLoader
             } else {
                 cl = clazz.getClassLoader();
             }
-            log.debug("      Loaded by " + cl.toString());
+            log.trace("      Loaded by " + cl.toString());
         }
         return clazz;
 
@@ -879,8 +859,8 @@ public abstract class WebappClassLoaderBase extends URLClassLoader
     @Override
     public URL findResource(final String name) {
 
-        if (log.isDebugEnabled()) {
-            log.debug("    findResource(" + name + ")");
+        if (log.isTraceEnabled()) {
+            log.trace("    findResource(" + name + ")");
         }
 
         checkStateForResourceLoading(name);
@@ -899,11 +879,11 @@ public abstract class WebappClassLoaderBase extends URLClassLoader
             url = super.findResource(name);
         }
 
-        if (log.isDebugEnabled()) {
+        if (log.isTraceEnabled()) {
             if (url != null) {
-                log.debug("    --> Returning '" + url.toString() + "'");
+                log.trace("    --> Returning '" + url.toString() + "'");
             } else {
-                log.debug("    --> Resource not found, returning null");
+                log.trace("    --> Resource not found, returning null");
             }
         }
         return url;
@@ -917,26 +897,16 @@ public abstract class WebappClassLoaderBase extends URLClassLoader
         ResourceEntry entry = new ResourceEntry();
         entry.lastModified = resource.getLastModified();
         synchronized (resourceEntries) {
-            if (!resourceEntries.containsKey(path)) {
-                resourceEntries.put(path, entry);
-            }
+            resourceEntries.putIfAbsent(path, entry);
         }
     }
 
 
-    /**
-     * Return an enumeration of <code>URLs</code> representing all of the resources with the given name. If no resources
-     * with this name are found, return an empty enumeration.
-     *
-     * @param name Name of the resources to be found
-     *
-     * @exception IOException if an input/output error occurs
-     */
     @Override
     public Enumeration<URL> findResources(String name) throws IOException {
 
-        if (log.isDebugEnabled()) {
-            log.debug("    findResources(" + name + ")");
+        if (log.isTraceEnabled()) {
+            log.trace("    findResources(" + name + ")");
         }
 
         checkStateForResourceLoading(name);
@@ -983,8 +953,8 @@ public abstract class WebappClassLoaderBase extends URLClassLoader
     @Override
     public URL getResource(String name) {
 
-        if (log.isDebugEnabled()) {
-            log.debug("getResource(" + name + ")");
+        if (log.isTraceEnabled()) {
+            log.trace("getResource(" + name + ")");
         }
 
         checkStateForResourceLoading(name);
@@ -995,13 +965,13 @@ public abstract class WebappClassLoaderBase extends URLClassLoader
 
         // (1) Delegate to parent if requested
         if (delegateFirst) {
-            if (log.isDebugEnabled()) {
-                log.debug("  Delegating to parent classloader " + parent);
+            if (log.isTraceEnabled()) {
+                log.trace("  Delegating to parent classloader " + parent);
             }
             url = parent.getResource(name);
             if (url != null) {
-                if (log.isDebugEnabled()) {
-                    log.debug("  --> Returning '" + url.toString() + "'");
+                if (log.isTraceEnabled()) {
+                    log.trace("  --> Returning '" + url.toString() + "'");
                 }
                 return url;
             }
@@ -1010,8 +980,8 @@ public abstract class WebappClassLoaderBase extends URLClassLoader
         // (2) Search local repositories
         url = findResource(name);
         if (url != null) {
-            if (log.isDebugEnabled()) {
-                log.debug("  --> Returning '" + url.toString() + "'");
+            if (log.isTraceEnabled()) {
+                log.trace("  --> Returning '" + url.toString() + "'");
             }
             return url;
         }
@@ -1020,16 +990,16 @@ public abstract class WebappClassLoaderBase extends URLClassLoader
         if (!delegateFirst) {
             url = parent.getResource(name);
             if (url != null) {
-                if (log.isDebugEnabled()) {
-                    log.debug("  --> Returning '" + url.toString() + "'");
+                if (log.isTraceEnabled()) {
+                    log.trace("  --> Returning '" + url.toString() + "'");
                 }
                 return url;
             }
         }
 
         // (4) Resource was not found
-        if (log.isDebugEnabled()) {
-            log.debug("  --> Resource not found, returning null");
+        if (log.isTraceEnabled()) {
+            log.trace("  --> Resource not found, returning null");
         }
         return null;
 
@@ -1064,8 +1034,8 @@ public abstract class WebappClassLoaderBase extends URLClassLoader
     @Override
     public InputStream getResourceAsStream(String name) {
 
-        if (log.isDebugEnabled()) {
-            log.debug("getResourceAsStream(" + name + ")");
+        if (log.isTraceEnabled()) {
+            log.trace("getResourceAsStream(" + name + ")");
         }
 
         checkStateForResourceLoading(name);
@@ -1076,21 +1046,21 @@ public abstract class WebappClassLoaderBase extends URLClassLoader
 
         // (1) Delegate to parent if requested
         if (delegateFirst) {
-            if (log.isDebugEnabled()) {
-                log.debug("  Delegating to parent classloader " + parent);
+            if (log.isTraceEnabled()) {
+                log.trace("  Delegating to parent classloader " + parent);
             }
             stream = parent.getResourceAsStream(name);
             if (stream != null) {
-                if (log.isDebugEnabled()) {
-                    log.debug("  --> Returning stream from parent");
+                if (log.isTraceEnabled()) {
+                    log.trace("  --> Returning stream from parent");
                 }
                 return stream;
             }
         }
 
         // (2) Search local repositories
-        if (log.isDebugEnabled()) {
-            log.debug("  Searching local repositories");
+        if (log.isTraceEnabled()) {
+            log.trace("  Searching local repositories");
         }
         String path = nameToPath(name);
         WebResource resource = resources.getClassLoaderResource(path);
@@ -1109,29 +1079,29 @@ public abstract class WebappClassLoaderBase extends URLClassLoader
             // Ignore
         }
         if (stream != null) {
-            if (log.isDebugEnabled()) {
-                log.debug("  --> Returning stream from local");
+            if (log.isTraceEnabled()) {
+                log.trace("  --> Returning stream from local");
             }
             return stream;
         }
 
         // (3) Delegate to parent unconditionally
         if (!delegateFirst) {
-            if (log.isDebugEnabled()) {
-                log.debug("  Delegating to parent classloader unconditionally " + parent);
+            if (log.isTraceEnabled()) {
+                log.trace("  Delegating to parent classloader unconditionally " + parent);
             }
             stream = parent.getResourceAsStream(name);
             if (stream != null) {
-                if (log.isDebugEnabled()) {
-                    log.debug("  --> Returning stream from parent");
+                if (log.isTraceEnabled()) {
+                    log.trace("  --> Returning stream from parent");
                 }
                 return stream;
             }
         }
 
         // (4) Resource was not found
-        if (log.isDebugEnabled()) {
-            log.debug("  --> Resource not found, returning null");
+        if (log.isTraceEnabled()) {
+            log.trace("  --> Resource not found, returning null");
         }
         return null;
     }
@@ -1173,9 +1143,9 @@ public abstract class WebappClassLoaderBase extends URLClassLoader
     @Override
     public Class<?> loadClass(String name, boolean resolve) throws ClassNotFoundException {
 
-        synchronized (getClassLoadingLock(name)) {
-            if (log.isDebugEnabled()) {
-                log.debug("loadClass(" + name + ", " + resolve + ")");
+        synchronized (JreCompat.isGraalAvailable() ? this : getClassLoadingLock(name)) {
+            if (log.isTraceEnabled()) {
+                log.trace("loadClass(" + name + ", " + resolve + ")");
             }
             Class<?> clazz = null;
 
@@ -1185,8 +1155,8 @@ public abstract class WebappClassLoaderBase extends URLClassLoader
             // (0) Check our previously loaded local class cache
             clazz = findLoadedClass0(name);
             if (clazz != null) {
-                if (log.isDebugEnabled()) {
-                    log.debug("  Returning class from cache");
+                if (log.isTraceEnabled()) {
+                    log.trace("  Returning class from cache");
                 }
                 if (resolve) {
                     resolveClass(clazz);
@@ -1195,10 +1165,10 @@ public abstract class WebappClassLoaderBase extends URLClassLoader
             }
 
             // (0.1) Check our previously loaded class cache
-            clazz = findLoadedClass(name);
+            clazz = JreCompat.isGraalAvailable() ? null : findLoadedClass(name);
             if (clazz != null) {
-                if (log.isDebugEnabled()) {
-                    log.debug("  Returning class from cache");
+                if (log.isTraceEnabled()) {
+                    log.trace("  Returning class from cache");
                 }
                 if (resolve) {
                     resolveClass(clazz);
@@ -1275,14 +1245,14 @@ public abstract class WebappClassLoaderBase extends URLClassLoader
 
             // (1) Delegate to our parent if requested
             if (delegateLoad) {
-                if (log.isDebugEnabled()) {
-                    log.debug("  Delegating to parent classloader1 " + parent);
+                if (log.isTraceEnabled()) {
+                    log.trace("  Delegating to parent classloader1 " + parent);
                 }
                 try {
                     clazz = Class.forName(name, false, parent);
                     if (clazz != null) {
-                        if (log.isDebugEnabled()) {
-                            log.debug("  Loading class from parent");
+                        if (log.isTraceEnabled()) {
+                            log.trace("  Loading class from parent");
                         }
                         if (resolve) {
                             resolveClass(clazz);
@@ -1295,14 +1265,14 @@ public abstract class WebappClassLoaderBase extends URLClassLoader
             }
 
             // (2) Search local repositories
-            if (log.isDebugEnabled()) {
-                log.debug("  Searching local repositories");
+            if (log.isTraceEnabled()) {
+                log.trace("  Searching local repositories");
             }
             try {
                 clazz = findClass(name);
                 if (clazz != null) {
-                    if (log.isDebugEnabled()) {
-                        log.debug("  Loading class from local repository");
+                    if (log.isTraceEnabled()) {
+                        log.trace("  Loading class from local repository");
                     }
                     if (resolve) {
                         resolveClass(clazz);
@@ -1315,14 +1285,14 @@ public abstract class WebappClassLoaderBase extends URLClassLoader
 
             // (3) Delegate to parent unconditionally
             if (!delegateLoad) {
-                if (log.isDebugEnabled()) {
-                    log.debug("  Delegating to parent classloader at end: " + parent);
+                if (log.isTraceEnabled()) {
+                    log.trace("  Delegating to parent classloader at end: " + parent);
                 }
                 try {
                     clazz = Class.forName(name, false, parent);
                     if (clazz != null) {
-                        if (log.isDebugEnabled()) {
-                            log.debug("  Loading class from parent");
+                        if (log.isTraceEnabled()) {
+                            log.trace("  Loading class from parent");
                         }
                         if (resolve) {
                             resolveClass(clazz);
@@ -1424,43 +1394,24 @@ public abstract class WebappClassLoaderBase extends URLClassLoader
     // ------------------------------------------------------ Lifecycle Methods
 
 
-    /**
-     * Add a lifecycle event listener to this component.
-     *
-     * @param listener The listener to add
-     */
     @Override
     public void addLifecycleListener(LifecycleListener listener) {
         // NOOP
     }
 
 
-    /**
-     * Get the lifecycle listeners associated with this lifecycle. If this Lifecycle has no listeners registered, a
-     * zero-length array is returned.
-     */
     @Override
     public LifecycleListener[] findLifecycleListeners() {
         return new LifecycleListener[0];
     }
 
 
-    /**
-     * Remove a lifecycle event listener from this component.
-     *
-     * @param listener The listener to remove
-     */
     @Override
     public void removeLifecycleListener(LifecycleListener listener) {
         // NOOP
     }
 
 
-    /**
-     * Obtain the current state of the source component.
-     *
-     * @return The current state of the source component.
-     */
     @Override
     public LifecycleState getState() {
         return state;
@@ -1479,11 +1430,6 @@ public abstract class WebappClassLoaderBase extends URLClassLoader
     }
 
 
-    /**
-     * Start the class loader.
-     *
-     * @exception LifecycleException if a lifecycle error occurs
-     */
     @Override
     public void start() throws LifecycleException {
 
@@ -1507,11 +1453,6 @@ public abstract class WebappClassLoaderBase extends URLClassLoader
     }
 
 
-    /**
-     * Stop the class loader.
-     *
-     * @exception LifecycleException if a lifecycle error occurs
-     */
     @Override
     public void stop() throws LifecycleException {
 
@@ -1579,19 +1520,21 @@ public abstract class WebappClassLoaderBase extends URLClassLoader
             }
         }
 
-        // De-register any remaining JDBC drivers
-        clearReferencesJdbc();
+        if (!JreCompat.isGraalAvailable()) {
+            // De-register any remaining JDBC drivers
+            clearReferencesJdbc();
+        }
 
         // Stop any threads the web application started
         clearReferencesThreads();
 
         // Clear any references retained in the serialization caches
-        if (clearReferencesObjectStreamClassCaches) {
+        if (clearReferencesObjectStreamClassCaches && !JreCompat.isGraalAvailable()) {
             clearReferencesObjectStreamClassCaches();
         }
 
         // Check for leaks triggered by ThreadLocals loaded by this class loader
-        if (clearReferencesThreadLocals) {
+        if (clearReferencesThreadLocals && !JreCompat.isGraalAvailable()) {
             checkThreadLocalsForLeaks();
         }
 
@@ -2035,8 +1978,8 @@ public abstract class WebappClassLoaderBase extends URLClassLoader
         return false;
     }
 
-    /*
-     * Get the set of current threads as an array.
+    /**
+     * @return the set of current threads as an array.
      */
     private Thread[] getThreads() {
         // Get the current thread group
@@ -2252,7 +2195,7 @@ public abstract class WebappClassLoaderBase extends URLClassLoader
             return clazz;
         }
 
-        synchronized (getClassLoadingLock(name)) {
+        synchronized (JreCompat.isGraalAvailable() ? this : getClassLoadingLock(name)) {
             clazz = entry.loadedClass;
             if (clazz != null) {
                 return clazz;
@@ -2346,6 +2289,17 @@ public abstract class WebappClassLoaderBase extends URLClassLoader
             } catch (UnsupportedClassVersionError ucve) {
                 throw new UnsupportedClassVersionError(
                         ucve.getLocalizedMessage() + " " + sm.getString("webappClassLoader.wrongVersion", name));
+            } catch (LinkageError e) {
+                // May be caused by the transformation also triggering loading of the class - BZ 68721
+                try {
+                    // Try and load the already defined class
+                    clazz = findLoadedClass0(name);
+                } catch (Throwable t) {
+                    // Not BZ 68721
+                    ExceptionUtils.handleThrowable(t);
+                    // Re-throw the original exception
+                    throw e;
+                }
             }
             entry.loadedClass = clazz;
         }
@@ -2522,21 +2476,6 @@ public abstract class WebappClassLoaderBase extends URLClassLoader
     }
 
 
-    /**
-     * Filter classes.
-     *
-     * @param name class name
-     *
-     * @return <code>true</code> if the class should be filtered
-     *
-     * @deprecated Use {@link #filter(String, boolean)} This will be removed in Tomcat 9
-     */
-    @Deprecated
-    protected boolean filter(String name) {
-        return filter(name, true) || filter(name, false);
-    }
-
-
     @Override
     protected void addURL(URL url) {
         super.addURL(url);
