@@ -20,7 +20,6 @@ import java.io.File;
 import java.io.FileNotFoundException;
 import java.io.FileOutputStream;
 import java.io.IOException;
-import java.io.InputStream;
 import java.io.OutputStreamWriter;
 import java.io.PrintWriter;
 import java.net.URI;
@@ -44,6 +43,7 @@ import org.apache.juli.logging.LogFactory;
 import org.apache.tomcat.util.digester.AbstractObjectCreationFactory;
 import org.apache.tomcat.util.digester.Digester;
 import org.apache.tomcat.util.file.ConfigFileLoader;
+import org.apache.tomcat.util.file.ConfigurationSource;
 import org.apache.tomcat.util.res.StringManager;
 import org.apache.tomcat.util.security.Escape;
 import org.xml.sax.Attributes;
@@ -160,9 +160,6 @@ public class MemoryUserDatabase implements UserDatabase {
 
     // ------------------------------------------------------------- Properties
 
-    /**
-     * @return the set of {@link Group}s defined in this user database.
-     */
     @Override
     public Iterator<Group> getGroups() {
         readLock.lock();
@@ -174,9 +171,6 @@ public class MemoryUserDatabase implements UserDatabase {
     }
 
 
-    /**
-     * @return the unique global identifier of this user database.
-     */
     @Override
     public String getId() {
         return this.id;
@@ -232,9 +226,6 @@ public class MemoryUserDatabase implements UserDatabase {
     }
 
 
-    /**
-     * @return the set of {@link Role}s defined in this user database.
-     */
     @Override
     public Iterator<Role> getRoles() {
         readLock.lock();
@@ -246,9 +237,6 @@ public class MemoryUserDatabase implements UserDatabase {
     }
 
 
-    /**
-     * @return the set of {@link User}s defined in this user database.
-     */
     @Override
     public Iterator<User> getUsers() {
         readLock.lock();
@@ -262,11 +250,6 @@ public class MemoryUserDatabase implements UserDatabase {
 
     // --------------------------------------------------------- Public Methods
 
-    /**
-     * Finalize access to this user database.
-     *
-     * @exception Exception if any exception is thrown during closing
-     */
     @Override
     public void close() throws Exception {
 
@@ -282,12 +265,6 @@ public class MemoryUserDatabase implements UserDatabase {
     }
 
 
-    /**
-     * Create and return a new {@link Group} defined in this user database.
-     *
-     * @param groupname The group name of the new group (must be unique)
-     * @param description The description of this group
-     */
     @Override
     public Group createGroup(String groupname, String description) {
         if (groupname == null || groupname.length() == 0) {
@@ -296,8 +273,7 @@ public class MemoryUserDatabase implements UserDatabase {
             throw new IllegalArgumentException(msg);
         }
 
-        @SuppressWarnings("deprecation")
-        Group group = new MemoryGroup(this, groupname, description);
+        Group group = new GenericGroup<>(this, groupname, description, null);
         readLock.lock();
         try {
             groups.put(group.getGroupname(), group);
@@ -308,12 +284,6 @@ public class MemoryUserDatabase implements UserDatabase {
     }
 
 
-    /**
-     * Create and return a new {@link Role} defined in this user database.
-     *
-     * @param rolename The role name of the new group (must be unique)
-     * @param description The description of this group
-     */
     @Override
     public Role createRole(String rolename, String description) {
         if (rolename == null || rolename.length() == 0) {
@@ -322,8 +292,7 @@ public class MemoryUserDatabase implements UserDatabase {
             throw new IllegalArgumentException(msg);
         }
 
-        @SuppressWarnings("deprecation")
-        Role role = new MemoryRole(this, rolename, description);
+        Role role = new GenericRole<>(this, rolename, description);
         readLock.lock();
         try {
             roles.put(role.getRolename(), role);
@@ -334,13 +303,6 @@ public class MemoryUserDatabase implements UserDatabase {
     }
 
 
-    /**
-     * Create and return a new {@link User} defined in this user database.
-     *
-     * @param username The logon username of the new user (must be unique)
-     * @param password The logon password of the new user
-     * @param fullName The full name of the new user
-     */
     @Override
     public User createUser(String username, String password, String fullName) {
 
@@ -350,8 +312,7 @@ public class MemoryUserDatabase implements UserDatabase {
             throw new IllegalArgumentException(msg);
         }
 
-        @SuppressWarnings("deprecation")
-        User user = new MemoryUser(this, username, password, fullName);
+        User user = new GenericUser<>(this, username, password, fullName, null, null);
         readLock.lock();
         try {
             users.put(user.getUsername(), user);
@@ -362,12 +323,6 @@ public class MemoryUserDatabase implements UserDatabase {
     }
 
 
-    /**
-     * Return the {@link Group} with the specified group name, if any; otherwise
-     * return <code>null</code>.
-     *
-     * @param groupname Name of the group to return
-     */
     @Override
     public Group findGroup(String groupname) {
         readLock.lock();
@@ -379,12 +334,6 @@ public class MemoryUserDatabase implements UserDatabase {
     }
 
 
-    /**
-     * Return the {@link Role} with the specified role name, if any; otherwise
-     * return <code>null</code>.
-     *
-     * @param rolename Name of the role to return
-     */
     @Override
     public Role findRole(String rolename) {
         readLock.lock();
@@ -396,12 +345,6 @@ public class MemoryUserDatabase implements UserDatabase {
     }
 
 
-    /**
-     * Return the {@link User} with the specified user name, if any; otherwise
-     * return <code>null</code>.
-     *
-     * @param username Name of the user to return
-     */
     @Override
     public User findUser(String username) {
         readLock.lock();
@@ -413,11 +356,6 @@ public class MemoryUserDatabase implements UserDatabase {
     }
 
 
-    /**
-     * Initialize access to this user database.
-     *
-     * @exception Exception if any exception is thrown during opening
-     */
     @Override
     public void open() throws Exception {
         writeLock.lock();
@@ -428,15 +366,8 @@ public class MemoryUserDatabase implements UserDatabase {
             roles.clear();
 
             String pathName = getPathname();
-            URI uri = ConfigFileLoader.getURI(pathName);
-            URLConnection uConn = null;
-
-            try {
-                URL url = uri.toURL();
-                uConn = url.openConnection();
-
-                InputStream is = uConn.getInputStream();
-                this.lastModified = uConn.getLastModified();
+            try (ConfigurationSource.Resource resource = ConfigFileLoader.getSource().getResource(pathName)) {
+                lastModified = resource.getLastModified();
 
                 // Construct a digester to read the XML input file
                 Digester digester = new Digester();
@@ -454,7 +385,7 @@ public class MemoryUserDatabase implements UserDatabase {
                         new MemoryUserCreationFactory(this), true);
 
                 // Parse the XML input to load this database
-                digester.parse(is);
+                digester.parse(resource.getInputStream());
             } catch (IOException ioe) {
                 log.error(sm.getString("memoryUserDatabase.fileNotFound", pathName));
             } catch (Exception e) {
@@ -463,15 +394,6 @@ public class MemoryUserDatabase implements UserDatabase {
                 groups.clear();
                 roles.clear();
                 throw e;
-            } finally {
-                if (uConn != null) {
-                    try {
-                        // Can't close a uConn directly. Have to do it like this.
-                        uConn.getInputStream().close();
-                    } catch (IOException ioe) {
-                        log.warn(sm.getString("memoryUserDatabase.fileClose", pathname), ioe);
-                    }
-                }
             }
         } finally {
             writeLock.unlock();
@@ -479,11 +401,6 @@ public class MemoryUserDatabase implements UserDatabase {
     }
 
 
-    /**
-     * Remove the specified {@link Group} from this user database.
-     *
-     * @param group The group to be removed
-     */
     @Override
     public void removeGroup(Group group) {
         readLock.lock();
@@ -500,11 +417,6 @@ public class MemoryUserDatabase implements UserDatabase {
     }
 
 
-    /**
-     * Remove the specified {@link Role} from this user database.
-     *
-     * @param role The role to be removed
-     */
     @Override
     public void removeRole(Role role) {
         readLock.lock();
@@ -526,11 +438,6 @@ public class MemoryUserDatabase implements UserDatabase {
     }
 
 
-    /**
-     * Remove the specified {@link User} from this user database.
-     *
-     * @param user The user to be removed
-     */
     @Override
     public void removeUser(User user) {
         readLock.lock();
@@ -574,12 +481,6 @@ public class MemoryUserDatabase implements UserDatabase {
     }
 
 
-    /**
-     * Save any updated information to the persistent storage location for this
-     * user database.
-     *
-     * @exception Exception if any exception is thrown during saving
-     */
     @Override
     public void save() throws Exception {
 
@@ -641,7 +542,7 @@ public class MemoryUserDatabase implements UserDatabase {
                         writer.print("\"");
                     }
                     writer.print(" roles=\"");
-                    for (Iterator<Role> roles=group.getRoles();roles.hasNext();) {
+                    for (Iterator<Role> roles=group.getRoles(); roles.hasNext();) {
                         Role role = roles.next();
                         writer.print(Escape.xml(role.getRolename()));
                         if(roles.hasNext()) {
@@ -665,7 +566,7 @@ public class MemoryUserDatabase implements UserDatabase {
                         writer.print("\"");
                     }
                     writer.print(" groups=\"");
-                    for (Iterator<Group> groups=user.getGroups();groups.hasNext();) {
+                    for (Iterator<Group> groups=user.getGroups(); groups.hasNext();) {
                         Group group = groups.next();
                         writer.print(Escape.xml(group.getGroupname()));
                         if(groups.hasNext()) {
@@ -673,7 +574,7 @@ public class MemoryUserDatabase implements UserDatabase {
                         }
                     }
                     writer.print("\" roles=\"");
-                    for (Iterator<Role> roles=user.getRoles();roles.hasNext();) {
+                    for (Iterator<Role> roles=user.getRoles(); roles.hasNext();) {
                         Role role = roles.next();
                         writer.print(Escape.xml(role.getRolename()));
                         if(roles.hasNext()) {
@@ -735,12 +636,13 @@ public class MemoryUserDatabase implements UserDatabase {
     }
 
 
+    @Override
     public void backgroundProcess() {
         if (!watchSource) {
             return;
         }
 
-        URI uri = ConfigFileLoader.getURI(getPathname());
+        URI uri = ConfigFileLoader.getSource().getURI(getPathname());
         URLConnection uConn = null;
         try {
             URL url = uri.toURL();
@@ -782,9 +684,6 @@ public class MemoryUserDatabase implements UserDatabase {
     }
 
 
-    /**
-     * Return a String representation of this UserDatabase.
-     */
     @Override
     public String toString() {
         StringBuilder sb = new StringBuilder("MemoryUserDatabase[id=");
@@ -808,7 +707,7 @@ public class MemoryUserDatabase implements UserDatabase {
  */
 class MemoryGroupCreationFactory extends AbstractObjectCreationFactory {
 
-    public MemoryGroupCreationFactory(MemoryUserDatabase database) {
+    MemoryGroupCreationFactory(MemoryUserDatabase database) {
         this.database = database;
     }
 
@@ -861,7 +760,7 @@ class MemoryGroupCreationFactory extends AbstractObjectCreationFactory {
  */
 class MemoryRoleCreationFactory extends AbstractObjectCreationFactory {
 
-    public MemoryRoleCreationFactory(MemoryUserDatabase database) {
+    MemoryRoleCreationFactory(MemoryUserDatabase database) {
         this.database = database;
     }
 
@@ -892,7 +791,7 @@ class MemoryRoleCreationFactory extends AbstractObjectCreationFactory {
  */
 class MemoryUserCreationFactory extends AbstractObjectCreationFactory {
 
-    public MemoryUserCreationFactory(MemoryUserDatabase database) {
+    MemoryUserCreationFactory(MemoryUserDatabase database) {
         this.database = database;
     }
 
